import random

  
class Broker:
    """
    A message passing system for binding the callbacks
    """

    def __init__(self):
        self.callback_map = {}
        
    def bind(self, msg, callback):
        self.callback_map.setdefault(msg, []).append(callback)

    def publish(self, msg, **kwargs):
        for callback in self.callback_map.get(msg, []):
            callback(**kwargs)

            

class GameObject:
    """
    GameObject has:
    x,y coordinates, positive integers, zero point is at upper left corner
    x, y sizes as positive integers, object rectangle is defined by x +/- x_size and  y +/- y_size
    collision detection/reaction
    """

    def __init__(self, x, y, x_size, y_size, obj_type="default", broker=Broker(), **kwargs):   # __init__ tekitab klassi eksemplari,
                                                # väljakutse mo=MovingObject(x,y,x_s,y_s)
                                                # NB! Väljakutsel jääb self argumendina ära 
        """ Initialize coordinates and sizes"""
        super().__init__(**kwargs)
        self.x = x # Tekitab ja väärtustab meie objektil atribuudi x
        self.y = y
        self.x_size = x_size
        self.y_size = y_size
        self.obj_type = obj_type
        self.broker = broker
        
    def collides(self, other):
        """ Detect if our size rect. overlaps with other GameObjects """
        x_overlaps = abs(self.x - other.x) < self.x_size + other.x_size
        y_overlaps = abs(self.y - other.y) < self.y_size + other.y_size
        return x_overlaps and y_overlaps

    def collision_react(self, other):
        """ React to collision with other GameObject """
        pass

    def destroy(self):
        self.broker.publish("Destroy", obj=self)


class Pellet(GameObject):

            
    def collision_react(self, other):
        if other.obj_type == "pacman":
            self.destroy()
            


class MovingObject(GameObject): # Kuna MovingObject on GameObjecti alamklass, siis
                                # võtab see kõik GameObjecti meetodid vaikimisi üle.                
    """
    MovingObject is a GameObject with additional movement methods and heading (dx, dy)
    It calls a callback function when moves.
    """

    def __init__(self, dx, dy, **kwargs):
        """ dx, dy: heading, change of x,y coordinates """
        super().__init__(**kwargs) # Ülemklassi meetodi väljakutse
        self.dx = dx
        self.dy = dy
 
    def set_heading(self, dx, dy):
        self.dx = dx
        self.dy = dy

    def tick(self):
        """ React to timer tick by moving """
        self.move(self.dx, self.dy)

    def move(self, dx=0, dy=0):
        """ Move the object by dx, dy """
        self.x += dx
        self.y += dy

        
class PacMan(MovingObject): # MovingObject on PacMani ülemklass
    """
    Represents the PacMan moving object.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.score = 0
        self.transformed_timer = 0

    def collision_react(self, other):
        if other.obj_type == "wall":
            self.move(-self.dx, -self.dy)
        elif other.obj_type == "ghost" and self.transformed_timer==0:
            self.lose_life()
        elif other.obj_type == "pellet":
            self.score += 1
            self.change_score()
        elif other.obj_type == "superpellet":
            self.score += 1
            self.change_score()
            self.transform()                  


    def tick(self):
        super().tick()    
        if self.transformed_timer > 0:
            self.transformed_timer -= 1
        if self.transformed_timer == 0:
            self.normalize()
        elif self.transformed_timer == 50:
            self.transform_warning()
        
                        
    def lose_life(self):
        self.broker.publish("Lose life", obj=self)

    def change_score(self):
        self.broker.publish("Change score", score=self.score, obj=self)

    def transform(self):
        self.transformed_timer = 200
        self.broker.publish("Transform", obj=self)

    def normalize(self):
        self.broker.publish("Normalize", obj=self)

    def transform_warning(self):
        self.broker.publish("Transform warning", obj=self)
       
    
    
class Ghost(MovingObject):
    """
    Represents the Ghost moving object.
    TBD: for each tick Ghost should have some small chance for changing the heading
    """
        
    def collision_react(self, other):
        if other.obj_type == "wall":
            self.move(-self.dx, -self.dy)
            self.set_rnd_heading()
        elif other.obj_type == "pacman" and other.transformed_timer>0:
            self.destroy()

    def set_rnd_heading(self):
        speed = abs(self.dx + self.dy)
        dx, dy = random.choice([(-speed, 0), (speed, 0), (0, -speed), (0, speed)])
        self.set_heading(dx, dy)
         
    def tick(self):
        # 1 juhul 100st, muuda juhuslikult suunda
        if random.randint(0, 50) == 1:
            self.set_rnd_heading()
        # Soorita ülemklassi tick() meetod
        super().tick()


class World(Broker):
    """
    A collection of all objects in the game.
    It forwards tick messages and detects collisions.
    It alse functions as a message broker for all its GameObjects.
    """

    def __init__(self, static_objs, moving_objs):
        """
        static_objs: list of static GameObjects
        moving_objs: list of MovingObjects
        """
        super().__init__()
        self.static_objs = static_objs
        self.moving_objs = moving_objs
        for obj in self.static_objs + self.moving_objs:
            obj.broker = self
        self.bind("Destroy", self.destroy)
        self.starting_pellets = self.pellet_count()


    def pellet_count(self):
        return len([x for x in self.static_objs if x.obj_type in ["pellet", "superpellet"]])

    def destroy(self, obj=None, **kwargs):
        """
        Register the destruction of the object
        """
        for objs in (self.static_objs, self.moving_objs):
            if obj in objs: objs.remove(obj)
        if self.pellet_count() <= 0:
            self.publish("Game won")

    def tick(self):
        """
        Forwards tick to all moving objects
        """
        for mo in self.moving_objs:
            mo.tick()
            # Kontrolli kokkupõrkeid
            for other_obj in self.static_objs + self.moving_objs:
                if mo != other_obj and mo.collides(other_obj):
                    # Saada sõnumid
                    mo.collision_react(other_obj)
                    other_obj.collision_react(mo)



class WorldBuilder:
    """
    Class for creating appropriate objects and adding them into the world
    """

    def __init__(self, x_centre=100, y_centre=100):
        self.static_objs = []
        self.moving_objs = []
        self.x_centre = x_centre
        self.y_centre = y_centre
        self.init_obj_maps()

    def init_obj_maps(self):
        """
        To be redefined by subclasses
        self.static_map and self.moving_map should be {obj_typ: Class} dicts  
        """
        self.static_map = {"wall": GameObject,
                           "pellet": Pellet,
                           "superpellet": Pellet}
        self.moving_map = {"ghost": Ghost,
                           "pacman": PacMan}

    def add_obj(self, obj_type, x=0, y=0, h_reflect=False, v_reflect=False, **kwargs):
        """
        Add a single object
        if h_reflect then reflect x horizontally, creating a copy
        if v_reflect then reflect y vertically, creating a copy
        if both h_reflect and v_reflect are true, create 3 copies.
        PacMan should not be reflected or added twice
        """
        if h_reflect:
            x_coords = [x, 2*self.x_centre - x] 
        else:
            x_coords = [x]
        if v_reflect:
            y_coords = [y, 2*self.y_centre - y] 
        else:
            y_coords = [y]
        for x in x_coords:
            for y in y_coords:
                if obj_type in self.static_map:
                    self.static_objs.append(self.static_map[obj_type](obj_type=obj_type, x=x, y=y, **kwargs))
                if obj_type in self.moving_map:
                    obj = self.moving_map[obj_type](obj_type=obj_type, x=x, y=y, **kwargs)
                    self.moving_objs.append(obj)
                    if obj_type == "pacman":
                        self.pacman = obj
            

    def add_obj_row(self, obj_type, x=0, y=0, x_step=0, y_step=0, n=10, **kwargs):
        """
        Create a row of n objects starting from x,y and adding x_step, y_step each step to 
        object coordinates.
        Both h_reflect and v_reflect are available
        """
        for i in range(n):
            self.add_obj(obj_type=obj_type, x = x + i*x_step, y = y + i*y_step, **kwargs)


    def get_world(self):
        return World(self.static_objs, self.moving_objs)

    def get_pacman(self):
        return self.pacman
  


def standard_world(wb):
    """
    Use WorldBuilder wb to build a standard Pacman Maze
    returns pacman, world
    """
    wb.x_centre = 220
    wb.y_centre = 220

    # Add Pacman
    wb.add_obj(y=100, x=220, x_size=7, y_size=7, dx=0, dy=0, obj_type="pacman")
    
    # Build edges
    wb.add_obj(x=220, y=20, x_size=200, y_size=10, obj_type="wall", v_reflect=True)
    wb.add_obj(y=220, x=20, x_size=10, y_size=200, obj_type="wall", h_reflect=True)
    
    # Build middle
    wb.add_obj(y=200, x=220, x_size=30, y_size=10, obj_type="wall")
    wb.add_obj(y=260, x=220, x_size=10, y_size=10, obj_type="wall")
    wb.add_obj(y=220, x=180, x_size=10, y_size=30, obj_type="wall", h_reflect=True)
    wb.add_obj(y=220, x=100, x_size=50, y_size=10, obj_type="wall", h_reflect=True)
    wb.add_obj(y=200, x=130, x_size=20, y_size=10, obj_type="wall", h_reflect=True)
    wb.add_obj_row(y=220, x=200, x_size=9, y_size=9, dx=3, dy=0, obj_type="ghost", x_step=20, n=3)
    
     
    # Add vertically reflected middle walls and pellets
    for x, y, xs, ys in [(220, 60, 10, 30), (220, 120, 50, 10), (220, 140, 10, 30)]:
        wb.add_obj(x=x, y=y, x_size=xs, y_size=ys, obj_type="wall", v_reflect=True)
    


    # Add 4x repeated walls and pellets
    for x, y, xs, ys in [(70, 70, 20, 20),
                         (150, 70, 40, 20),
                         (70, 120, 20, 10),
                         (60, 170, 30, 20),
                         (130, 150, 20, 40),
                         (170, 160, 20, 10)]:
        wb.add_obj(x=x, y=y, x_size=xs, y_size=ys, obj_type="wall", v_reflect=True, h_reflect=True)
    # Add pellets
    wb.add_obj(x=40, y=120, x_size=5, y_size=5, obj_type="superpellet", v_reflect=True, h_reflect=True)
    for x, y, x_step, y_step, n in [(160, 180, 20, 0, 3),
                                    (40, 40, 20, 0, 9),
                                    (40, 100, 20, 0, 9),
                                    (40, 60, 0, 20, 2),
                                    (100, 60, 0, 20, 2),
                                    (200, 60, 0, 20, 2),
                                    (100, 120, 0, 20, 5)]:
        wb.add_obj_row(x=x, y=y, x_size=3, y_size=3, obj_type="pellet",
                       x_step=x_step, y_step=y_step, n=n, v_reflect=True, h_reflect=True)
    
    # Return Pacman, World
    return wb.get_pacman(), wb.get_world()






if __name__ == "__main__":
    wb = WorldBuilder(x_centre=100, y_centre=100)
    wb.add_obj_row(x=100, y=100, x_step=0, y_step=20, n= 10, x_size=10, y_size=10, obj_type="pellet")
    wb.add_obj(x=115, y=115, x_size=10, y_size=10, dx=0, dy=10, obj_type="pacman")
    wb.add_obj(x=115, y=115, x_size=10, y_size=10, dx=0, dy=-10, obj_type="ghost")
    #print(mo1.collides(mo2), mo2.collides(mo1))
    #p2 = Pellet(x=115, y=200, x_size=10, y_size=10, obj_type="pellet")
    wb.add_obj(x=115, y=25, x_size=30, y_size=5, h_reflect=True, obj_type="wall")
    def eaten(obj): print("eaten!", obj)
    #w = World([p1, p2, w1], [pm, g])
    w = wb.get_world()
    w.bind("Destroy", eaten)
    #print(mo3.collides(mo2), mo2.collides(mo3))
    for step in range(14):
        w.tick()
        #mo2.tick()
        #mo2.set_heading(0, 10)
        #print("step", step,  pm.x, pm.y, g.x, g.y)
        
    
        
